//
//  Unicode.Scalar+Information.swift
//
//  CotEditor
//  https://coteditor.com
//
//  Created by 1024jp on 2016-07-26.
//
//  ---------------------------------------------------------------------------
//
//  © 2015-2020 1024jp
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  https://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//

extension Unicode.Scalar {
    
    /// Code point string in format like `U+000F`.
    var codePoint: String {
        
        return String(format: "U+%04tX", self.value)
    }
    
    
    /// Code point pair in UTF-16 surrogate pair.
    var surrogateCodePoints: [String]? {
        
        guard self.isSurrogatePair else { return nil }
        
        return [String(format: "U+%04X", UTF16.leadSurrogate(self)),
                String(format: "U+%04X", UTF16.trailSurrogate(self))]
    }
    
    
    /// Boolean value indicating whether character becomes a surrogate pair in UTF-16.
    var isSurrogatePair: Bool {
        
        return (UTF16.width(self) == 2)
    }
    
    
    /// Unicode name.
    var name: String? {
        
        return self.properties.nameAlias
            ?? self.properties.name
            ?? self.controlCharacterName  // get control character name from special table
    }
    
    
    /// Unicode block name.
    var blockName: String? {
        
        return self.value.blockName
    }
    
    
    /// Localized and sanitized unicode block name.
    var localizedBlockName: String? {
        
        // -> This is actually a dirty workaround to make the block name the same as the Apple's block naming rule.
        //    Otherwise, we cannot localize block names correctly. (2015-11)
        
        return self.blockName?
            .replacingOccurrences(of: " ([A-Z])$", with: "-$1", options: .regularExpression)
            .replacingOccurrences(of: "Supplement", with: "Supp.")
            .replacingOccurrences(of: "Description", with: "Desc.")
            .localized(tableName: "Unicode")
    }
    
}



// MARK: -

extension UTF32.CodeUnit {
    
    /// Return Unicode name.
    ///
    /// Implemented at UTF32.CodeUnit level in order to cover single surrogate characters
    /// that are not allowed by Unicode.Scalar.
    var unicodeName: String? {
        
        if let name = Unicode.Scalar(self)?.name {
            return name
        }
        
        if let codeUnit = UTF16.CodeUnit(exactly: self) {
            if UTF16.isLeadSurrogate(codeUnit) {
                return "<lead surrogate-" + String(format: "%04X", self) + ">"
            }
            if UTF16.isTrailSurrogate(codeUnit) {
                return "<tail surrogate-" + String(format: "%04X", self) + ">"
            }
        }
        
        return nil
    }
    
}
